#
# PST lib - Postscript DSC chewing
#   Simple but hopefully robust enough
# 
# (c) 2002  Tomas Valla   <valla@atrey.karlin.mff.cuni.cz>
# 
# This software may be used and distributed under the terms
# of the GNU General Public License. See COPYING for details.
#

package DSC;
use strict;
use IO::File;

sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = {};

	$self->{VERBOSE} = 0;  #  0=quiet, 1=normal, 2=some dumping, 3=superinfo
	$self->{GLOBALS} = {};   # header comments
	$self->{HEADER_START} = undef;   # positions of various elements
	$self->{HEADER_END} = undef;
	$self->{PROLOG_START} = undef;
	$self->{PROLOG_END} = undef;
	$self->{SETUP_START} = undef;
	$self->{SETUP_END} = undef;
	$self->{DEFAULTS_END} = undef;
	$self->{BODY_START} = undef;
	$self->{PAGE_STARTS} = {};
	$self->{PAGE_SETUP_STARTS} = {};
	$self->{PAGE_SETUP_ENDS} = {};
	$self->{PAGE_TRAILER_STARTS} = {};
	$self->{PAGE_NUMBERS} = {};
	$self->{PAGE_ATTRS} = {};      # single page comments
	$self->{NUM_PAGES} = undef;
	$self->{TRAILER_START} = undef;
	$self->{END_START} = undef;
	$self->{FIRST_RESOURCE_START} = undef;
	$self->{PROCSETS_SEEN} = {};  # which ProcSets are present in document
	$self->{PAGE_COUNT} = 0;   # pagenumber for write_page()

	return bless ($self, $class);
}


# set verbosity level,
sub verbose {
	my $self = shift;
	$self->{VERBOSE} = shift;
}


# open the file given as first argument for reading
sub open {
	my $self = shift;
	my $name = shift;

	$self->{NAME} = $name;
	$self->{SFH} = new IO::File;

	return $self->{SFH}->open("<$name");
}


# destructor - close open files
sub DESTROY {
	my $self = shift;

	defined $self->{SFH} && $self->{SFH}->close;
	defined $self->{DFH} && $self->{DFH}->close;
}


# main DSC parsing routine
# return error string, "" if all ok
sub parse {
	my $self = shift;
	my $start;

	print "Parsing ", $self->{NAME}, "\n" if $self->{VERBOSE};

	# establish pipe to psscan
	my $t = new IO::File;
	$t->open("psscan <" . $self->{NAME} . "|") || return "Unable to open pipe";
	$_ = <$t>; defined and chomp;
	(defined and /^(\d)+ %!PS/) || return "Not a PostScript file";

	$self->{HEADER_START} = $1;

	# process header
	while (<$t>) {
		print if $self->{VERBOSE} >= 3;
		/^(\d+)+ %%(\w+:?)\s*(.*)\n$/ || return "Parse error";
		($2 eq "EndComments") && do {
			$self->{HEADER_END} = $1;
			last;
		};
		($2 eq "EOF") && return "Premature EOF in document header";
		defined $self->{GLOBALS}{$2} or $self->{GLOBALS}{$2} = $3;
	}

	my $page = 0;
	my $sect = 0;  # 0=hdr,1=setup,2=hdrseensetup,3=pages,4=trailer
	my $psetup = 0;
	while (<$t>) {
		print if $self->{VERBOSE} >= 3;
		/^(\d+)+ %%(\w+:?)\s*(.*)\n$/ || return "Parse error";
		my $pos = $1;
		if ($2 eq "BeginSetup") {
			!$sect or return "%%BeginSetup: Sequencing error";
			$sect = 1;
			!defined($self->{SETUP_START}) or return "Multiple setup sections";
			$self->{SETUP_START} = $pos;
		} elsif ($2 eq "EndSetup") {
			$sect == 1 or return "Misplaced %%EndSetup";
			$sect = 2;
			$self->{SETUP_END} = $pos;
		} elsif ($2 eq "Page:") {
			$sect != 1 && $sect != 4 or return "Misplaced %%Page";
			$sect = 3;
			$3 =~ /^(.*)\s+(\d+)$/ or return "%%Page out of sequence";
			++$page == $2 or return "Misordered %%Page";
			$self->{PAGE_STARTS}{$page} = $pos;
			$self->{PAGE_NUMBERS}{$page} = parse_ps_string($1);
			$self->{PAGE_ATTRS}{$page} = {};
			$psetup == 1 && return "Missing %%EndPageSetup";
			$psetup = 0;
		} elsif ($2 eq "BeginPageSetup") {
			$sect == 3 or return "Misplaced %%BeginPageSetup";
			$self->{PAGE_SETUP_STARTS}{$page} = $pos;
			$psetup = 1;
		} elsif ($2 eq "EndPageSetup") {
			$sect == 3 or return "Misplaced %%EndPageSetup";
			$self->{PAGE_SETUP_ENDS}{$page} = $pos;
			$psetup = 0;
		} elsif ($2 eq "PageTrailer") {
			$sect == 3 or return "Misplaced %%PageTrailer";
			$self->{PAGE_TRAILER_STARTS}{$page} = $pos;
			$psetup = -1;
		} elsif ($2 eq "Trailer") {
			($sect == 0 || $sect == 2 || $sect == 3) or return "Misplaced %%Trailer";
			$sect = 4;
			$self->{TRAILER_START} = $pos;
		} elsif ($2 eq "EOF") {
			$self->{END_START} = $pos;
			last;
		} elsif ($2 eq "BeginProcSet:") {
			$sect == 0 or next;
			defined $self->{FIRST_RESOURCE_START} or $self->{FIRST_RESOURCE_START} = $pos;
			$self->{PROCSETS_SEEN}{$3} = "";
		} elsif ($2 eq "BeginResource:") {
			$sect == 0 or next;
			defined $self->{FIRST_RESOURCE_START} or $self->{FIRST_RESOURCE_START} = $pos;
			if ($3 =~ /^procset\s+(\w+)/) { $self->{PROCSETS_SEEN}{$1} = ""; }
		} elsif ($2 eq "EndDefaults" && !$sect) {
			$self->{DEFAULTS_END} = $pos;
		} elsif ($2 eq "BeginProlog" && !$sect) {
			$self->{PROLOG_START} = $pos;
		} elsif ($2 eq "EndProlog" && !$sect) {
			$self->{PROLOG_END} = $pos;
		} elsif ($sect == 4) {
			$self->{GLOBALS}{$2} = $3;
		} elsif ($page > 0) {  # other comments
			if ($psetup < 0 ||
			    $psetup > 0 && !defined $self->{PAGE_ATTRS}{$2}) {
				$self->{PAGE_ATTRS}{$page}{$2} = $3;
			}
		}
	}
	defined $self->{END_START} or return "Incomplete document";
	$t->close;

	$self->{NUM_PAGES} = $page;
	$page || return "Document contains no pages";
	my $k;
	foreach $k (keys %{$self->{GLOBALS}}) {
		($self->{GLOBALS}{$k} eq "(atend)") &&
			return "$k deferred at end, but not defined there";
	}
	$self->{BODY_START} = $self->{PAGE_STARTS}{1};

	$self->dump if $self->{VERBOSE} >= 2;
	return "";
}

sub d {
	my $x = shift;
	return defined $x ? $x : "???";
}

# print some info...
sub dump {
	my $self = shift;
	print "--- Document information ---\n";
	print "Name:\t\t", d($self->{NAME}), "\n";
	foreach my $i (keys %{$self->{GLOBALS}}) {
		print "\t$i ", $self->{GLOBALS}{$i}, "\n";
	}
	print "HdrStart:\t", d($self->{HEADER_START}), "\n";
	print "PrologStart:\t", d($self->{HEADER_END}), "\n";
	print "Setup:\t\t", d($self->{SETUP_START}), "...", d($self->{SETUP_END}), "\n";
	print "BodyStart:\t", d($self->{BODY_START}), "\n";
	if ($self->{VERBOSE} >= 3) {
		for (my $i=1; $i <= $self->{NUM_PAGES}; $i++) {
			print "Page $i: at ", $self->{PAGE_STARTS}{$i};
			defined $self->{PAGE_SETUP_STARTS}{$i} and
				print " setup ", d($self->{PAGE_SETUP_STARTS}{$i}), "...",
					d($self->{PAGE_SETUP_ENDS}{$i});
			defined $self->{PAGE_TRAILER_STARTS}{$i} and
				print " trailer ", d($self->{PAGE_TRAILER_STARTS}{$i});
			print " name ", d($self->{PAGE_NUMBERS}{$i}), "\n";
		}
	} else {
		print "Pages:\t", $self->{NUM_PAGES}, "\n";
	}
	print "TrailerStart:\t", d($self->{TRAILER_START}), "\n";
	print "END:\t\t", d($self->{END_START}), "\n";
}

sub parse_ps_string {
	my $x = shift;
	return ($x =~ /^\((.*)\)$/) ? $1 : $x;
}

# output file init
sub write_start {
	my $self = shift;
	my $fd = shift;
	return $self->{DFH} = $fd;
}

sub write_header {
	my $self = shift;
	my $h = $self->{DFH};
	print $h "%!PS-Adobe-3.0\n";
	foreach my $z (keys %{$self->{GLOBALS}}) {
		my $val = $self->{GLOBALS}{$z};
		print $h "%%$z", ($val eq "") ? "" : " ", $val, "\n";
	}
	print $h "%%EndComments\n";
}

sub write_prolog {
	my $self = shift;
	my $s = $self->{SFH};
	my $h = $self->{DFH};

	seek $s, $self->{HEADER_END}, 0;
	# skip %%EndComments line, if present
	if (<$s> !~ /^%%EndComments(\s|$)/) { seek $s, $self->{HEADER_END}, 0; }

	if (defined $self->{PROCSETS_SEEN}{"PSTprocset"}) {
		# already there
	} elsif (defined $self->{FIRST_RESOURCE_START}) {
		copy_until($s, $h, $self->{FIRST_RESOURCE_START});
		write_procset($h);
	} elsif (defined $self->{PROLOG_START}) {
		copy_until($s, $h, $self->{PROLOG_START});
		my $ll = <$s>;
		print $h $ll;
		write_procset($h);
	} else {
		if (defined $self->{DEFAULTS_END}) {
			copy_until($s, $h, $self->{DEFAULTS_END});
			my $ll = <$s>;
			print $h $ll;
		}
		print $h "%%BeginProlog\n";
		write_procset($h);
		defined $self->{PROLOG_END} || print $h "%%EndProlog\n";
	}
	copy_until($s, $h, $self->{BODY_START});
	if (not defined $self->{PROCSETS_SEEN}{"PSTprocset"}) {
		print $h
			"userdict/PSTxform PSTmatrix matrix currentmatrix\n",
			"matrix invertmatrix matrix concatmatrix\n",
			"matrix invertmatrix put\n";
	}
}

sub copy_until {
	my $s = shift;
	my $h = shift;
	my $until = shift;
	while (tell $s < $until) {
		my $l = $until - tell $s;
		if ($l > 4096) { $l = 4096; }
		my $buf;
		read($s, $buf, $l) == $l or die "Error while reading input file";
		print $h $buf;
	}
}

sub write_procset {
	my $h = shift;
	# print $h "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% HERE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
	my @procset = (
		"%%BeginProcSet: PSTprocset",
		"userdict begin",
		"[/showpage/erasepage/copypage]{dup where{pop dup load",
		" type/operatortype eq{1 array cvx dup 0 3 index cvx put",
		" bind def}{pop}ifelse}{pop}ifelse}forall",
		"[/letter/legal/executivepage/a4/a4small/b5/com10envelope", 
		" /monarchenvelope/c5envelope/dlenvelope/lettersmall/note",
		" /folio/quarto/a5]{dup where{dup wcheck{exch{}put}",
		" {pop{}def}ifelse}{pop}ifelse}forall",
		"/setpagedevice {pop}bind 1 index where{dup wcheck{3 1 roll put}",
		" {pop def}ifelse}{def}ifelse",
		"/PSTmatrix matrix currentmatrix def",
		"/PSTxform matrix def/PSTclip{clippath}def",
		"/defaultmatrix{PSTmatrix exch PSTxform exch concatmatrix}bind def",
		"/initmatrix{matrix defaultmatrix setmatrix}bind def",
		"/initclip[{matrix currentmatrix PSTmatrix setmatrix",
		" [{currentpoint}stopped{\$error/newerror false put{newpath}}",
		" {/newpath cvx 3 1 roll/moveto cvx 4 array astore cvx}ifelse]",
		" {[/newpath cvx{/moveto cvx}{/lineto cvx}",
		" {/curveto cvx}{/closepath cvx}pathforall]cvx exch pop}",
		" stopped{\$error/errorname get/invalidaccess eq{cleartomark",
		" \$error/newerror false put cvx exec}{stop}ifelse}if}bind aload pop",
		" /initclip dup load dup type dup/operatortype eq{pop exch pop}",
		" {dup/arraytype eq exch/packedarraytype eq or",
		"  {dup xcheck{exch pop aload pop}{pop cvx}ifelse}",
		"  {pop cvx}ifelse}ifelse",
		" {newpath PSTclip clip newpath exec setmatrix} bind aload pop]cvx def",
		"/initgraphics{initmatrix newpath initclip 1 setlinewidth",
		" 0 setlinecap 0 setlinejoin []0 setdash 0 setgray",
		" 10 setmiterlimit}bind def",
		#"/setmatrix {matrix currentmatrix concat} bind def",
		"end",
		"%%EndProcSet",
	);
	for my $line (@procset) {
		print $h $line,"\n";
	}
}

sub write_page {
	my $self = shift;
	my $page = shift;
	#my $page_count = ++$self->{PAGE_COUNT};

	$self->write_page_header($page);
	$self->write_page_body($page);
}

sub write_page_header {
	my $self = shift;
	my $page = shift;
	my $aspage = ++ $self->{PAGE_COUNT};
	my $h = $self->{DFH};

	print $h "%%Page: ",defined $self->{PAGE_NUMBERS}{$page} ? $self->{PAGE_NUMBERS}{$page} : $aspage, 
		" ", defined $aspage ? $aspage : $page, "\n";
	#foreach my $k (keys %{$self->{PAGE_ATTRS}{$page}}){
	#	my $val=$self->{PAGE_ATTRS}{$page}{$k};
	#	print $h "%%$k", ($val eq "") ? "" : " ",$val,"\n";
	#}
}

sub write_page_body {
	my $self = shift;
	my $page = shift;
	my $aspage = $self->{PAGE_COUNT};
	my $s = $self->{SFH};
	my $h = $self->{DFH};

	seek $s, $self->{PAGE_STARTS}{$page}, 0;
	# print $h "%%Page: ",$self->{PAGE_NUMBERS}{$page}," ", 
	#	defined $aspage ? $aspage : $page, "\n";
	if(<$s> !~ /^%%Page:/) { seek $s, $self->{PAGE_STARTS}{$page}, 0; }
	# foreach my $k (keys %{$self->{PAGE_ATTRS}{$page}}){
	#	my $val=$self->{PAGE_ATTRS}{$page}{$k};
	#	print $h "%%$k", ($val eq "") ? "" : " ",$val,"\n";
	# }
	if(defined $self->{PAGE_SETUP_STARTS}{$page}){
		copy_until($s,$h,$self->{PAGE_SETUP_STARTS}{$page}); # should be empty, but DSC conformance is very mild ...
		print $h "%%BeginPageSetup\n";
		seek $s,$self->{PAGE_SETUP_STARTS}{$page},0;
		if (<$s> !~ /^%%BeginPageSetup(\s|$)/) { seek $s, $self->{PAGE_SETUP_STARTS}{$page}, 0; }
		copy_until($s,$h,$self->{PAGE_SETUP_ENDS}{$page});
		print $h "%%EndPageSetup\n";
		if (<$s> !~ /^%%EndPageSetup(\s|$)/) { seek $s, $self->{PAGE_SETUP_ENDS}{$page},0; }
	}
	if(defined $self->{PAGE_TRAILER_STARTS}{$page}){
		copy_until($s,$h,$self->{PAGE_TRAILER_STARTS}{$page});
		print $h "%%PageTrailer\n";
		if(<$s> !~ /^%%PageTrailer(\s|$)/) { seek $s, $self->{PAGE_TRAILER_STARTS}{$page},0; }
	}
	if(defined $self->{PAGE_STARTS}{$page+1}){
		copy_until($s,$h,$self->{PAGE_STARTS}{$page+1});
	}elsif(defined $self->{TRAILER_START}){
		copy_until($s,$h,$self->{TRAILER_START});
	}else{           # Only %%EOF should remain
		copy_until($s,$h,$self->{END_START});
	}
}

sub write_epilog {
	my $self = shift;
	my $h = $self->{DFH};
	my $s = $self->{SFH};
	if(defined $self->{TRAILER_START}){
		print $h "%%Trailer\n";
		seek $s,$self->{TRAILER_START},0;
		if(<$s> !~ /^%%Trailer(\s|$)/) { seek $s,$self->{TRAILER_START},0; }
		copy_until($s,$h,$self->{END_START});
	}
	print $h "%%EOF\n";
}

# We do not perform any initialization
1;
